"""
The most frequently used (can serve as baseline) CNN family of physiological signal processing,
whose performance however seems exceeded by newer networks
"""

from copy import deepcopy
from itertools import repeat
from numbers import Real
from typing import NoReturn, Optional, Sequence, Union

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from ...cfg import CFG, DEFAULTS
from ...models._nets import (  # noqa: F401
    Activations,
    Conv_Bn_Activation,
    DownSample,
    MultiConv,
    SpaceToDepth,
    ZeroPadding,
    make_attention_layer,
)
from ...utils.misc import dict_to_str, add_docstring
from ...utils.utils_nn import (
    SizeMixin,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)

if DEFAULTS.torch_dtype == torch.float64:
    torch.set_default_tensor_type(torch.DoubleTensor)


__all__ = [
    "ResNet",
    "ResNetBasicBlock",
    "ResNetBottleNeck",
]


class ResNetBasicBlock(nn.Module, SizeMixin):
    """
    building blocks for `ResNet`, as implemented in ref. [2] of `ResNet`

    """

    __DEBUG__ = False
    __name__ = "ResNetBasicBlock"
    expansion = 1  # not used

    def __init__(
        self,
        in_channels: int,
        num_filters: int,
        filter_length: int,
        subsample_length: int,
        groups: int = 1,
        dilation: int = 1,
        attn: Optional[dict] = None,
        **config,
    ) -> NoReturn:
        """

        Parameters
        ----------
        in_channels: int,
            number of features (channels) of the input
        num_filters: int,
            number of filters for the convolutional layers
        filter_length: int,
            length (size) of the filter kernels
        subsample_lengths: int,
            subsample length,
            including pool size for short cut, and stride for the top convolutional layer
        groups: int, default 1,
            pattern of connections between inputs and outputs,
            for more details, ref. `nn.Conv1d`
        dilation: int, default 1,
            not used
        attn: dict, optional,
            attention mechanism for the neck conv layer,
            if None, no attention mechanism is used,
            keys:
                "name": str, can be "se", "gc", "nl" (alias "nonlocal", "non-local"), etc.
                "pos": int, position of the attention mechanism,
                other keys are specific to the attention mechanism
        config: dict,
            other hyper-parameters, including
            increase channel method, subsample method, dropouts,
            activation choices, weight initializer, and short cut patterns, etc.

        """
        super().__init__()
        if dilation > 1:
            raise NotImplementedError(f"Dilation > 1 not supported in {self.__name__}")
        self.__num_convs = 2
        self.__in_channels = in_channels
        self.__out_channels = num_filters
        self.__kernel_size = filter_length
        self.__down_scale = subsample_length
        self.__stride = subsample_length
        self.__groups = groups
        self.config = CFG(deepcopy(config))
        if self.__DEBUG__:
            print(
                f"configuration of {self.__name__} is as follows\n{dict_to_str(self.config)}"
            )

        if (
            self.config.increase_channels_method.lower() == "zero_padding"
            and self.__groups != 1
        ):
            raise ValueError(
                "zero padding for increasing channels can not be used with groups != 1"
            )

        self.__attn = attn
        if self.__attn:
            self.__attn = CFG(self.__attn)

        self.__increase_channels = self.__out_channels > self.__in_channels
        self.shortcut = self._make_shortcut_layer()

        self.main_stream = nn.Sequential()
        conv_in_channels = self.__in_channels
        for i in range(self.__num_convs):
            conv_activation = (
                self.config.activation if i < self.__num_convs - 1 else None
            )
            if self.__attn and self.__attn["pos"] == i:
                self.main_stream.add_module(
                    self.__attn["name"],
                    make_attention_layer(conv_in_channels, **self.__attn),
                )
            self.main_stream.add_module(
                f"cba_{i}",
                Conv_Bn_Activation(
                    in_channels=conv_in_channels,
                    out_channels=self.__out_channels,
                    kernel_size=self.__kernel_size,
                    stride=(self.__stride if i == 0 else 1),
                    groups=self.__groups,
                    norm=True,
                    activation=conv_activation,
                    kw_activation=self.config.kw_activation,
                    kernel_initializer=self.config.kernel_initializer,
                    kw_initializer=self.config.kw_initializer,
                    bias=self.config.bias,
                    conv_type=self.config.get("conv_type", None),
                ),
            )
            if self.config.get("dropout", 0) > 0 and i < self.__num_convs - 1:
                self.main_stream.add_module(
                    f"dropout_{i}", nn.Dropout(self.config.dropout)
                )
            conv_in_channels = self.__out_channels
        if self.__attn and self.__attn["pos"] == -1:
            self.main_stream.add_module(
                self.__attn["name"],
                make_attention_layer(conv_in_channels, **self.__attn),
            )

        if isinstance(self.config.activation, str):
            self.out_activation = Activations[self.config.activation.lower()](
                **self.config.kw_activation
            )
        else:
            self.out_activation = self.config.activation(**self.config.kw_activation)

        if self.config.get("dropout", 0) > 0:
            self.out_dropout = nn.Dropout(self.config.dropout)
        else:
            self.out_dropout = None

    def _make_shortcut_layer(self) -> Union[nn.Module, None]:
        """ """
        if self.__DEBUG__:
            print(
                f"down_scale = {self.__down_scale}, increase_channels = {self.__increase_channels}"
            )
        if self.__down_scale > 1 or self.__increase_channels:
            if self.config.increase_channels_method.lower() == "conv":
                shortcut = DownSample(
                    down_scale=self.__down_scale,
                    in_channels=self.__in_channels,
                    out_channels=self.__out_channels,
                    groups=self.__groups,
                    norm=True,
                    mode=self.config.subsample_mode,
                    filt_size=self.config.get("filt_size", 3),  # for blur pool
                )
            elif self.config.increase_channels_method.lower() == "zero_padding":
                batch_norm = (
                    False if self.config.subsample_mode.lower() != "conv" else True
                )
                shortcut = nn.Sequential(
                    DownSample(
                        down_scale=self.__down_scale,
                        in_channels=self.__in_channels,
                        out_channels=self.__in_channels,
                        norm=batch_norm,
                        mode=self.config.subsample_mode,
                        filt_size=self.config.get("filt_size", 3),  # for blur pool
                    ),
                    ZeroPadding(self.__in_channels, self.__out_channels),
                )
        else:
            shortcut = None
        return shortcut

    def forward(self, input: Tensor) -> Tensor:
        """

        Parameters
        ----------
        input: Tensor,
            of shape (batch_size, n_channels, seq_len)

        Returns
        -------
        out: Tensor,
            of shape (batch_size, n_channels, seq_len)

        """
        identity = input

        out = self.main_stream(input)

        if self.shortcut is not None:
            identity = self.shortcut(input)

        if identity.shape[-1] < out.shape[-1]:
            # downsampling using "avg" or "max" adopts no padding
            # hence shape[-1] of identity might be smaller by 1 than shape[-1] of out
            diff_sig_len = out.shape[-1] - identity.shape[-1]
            identity = F.pad(
                identity, [diff_sig_len // 2, diff_sig_len - diff_sig_len // 2]
            )

        out += identity
        out = self.out_activation(out)

        if self.out_dropout is not None:
            out = self.out_dropout(out)

        return out

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """

        Parameters
        ----------
        seq_len: int,
            length of the 1d sequence
        batch_size: int, optional,
            the batch size, can be None

        Returns
        -------
        output_shape: sequence,
            the output shape of this block, given `seq_len` and `batch_size`

        """
        _seq_len = seq_len
        for module in self.main_stream:
            if isinstance(module, nn.Dropout):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class ResNetBottleNeck(nn.Module, SizeMixin):
    """

    bottle neck blocks for `ResNet`, as implemented in ref. [2] of `ResNet`,
    as for 1D ECG, should be of the "baby-giant-baby" pattern?

    """

    __DEBUG__ = False
    __name__ = "ResNetBottleNeck"
    expansion = 4
    __DEFAULT_BASE_WIDTH__ = 12 * 4

    def __init__(
        self,
        in_channels: int,
        num_filters: int,
        filter_length: int,
        subsample_length: int,
        groups: int = 1,
        dilation: int = 1,
        base_width: int = 12 * 4,
        base_groups: int = 1,
        base_filter_length: int = 1,
        attn: Optional[dict] = None,
        **config,
    ) -> NoReturn:
        """

        Parameters
        ----------
        in_channels: int,
            number of features (channels) of the input
        num_filters: sequence of int,
            number of filters for the neck conv layer
        filter_length: int,
            lengths (sizes) of the filter kernels for the neck conv layer
        subsample_length: int,
            subsample length,
            including pool size for short cut,
            and stride for the (top or neck) conv layer
        groups: int, default 1,
            pattern of connections between inputs and outputs of the neck conv layer,
            for more details, ref. `nn.Conv1d`
        dilation: int, default 1,
            dilation of the conv layers
        base_width: int, default 12*4,
            number of filters per group for the neck conv layer
            usually number of filters of the initial conv layer of the whole ResNet
        base_groups: int, default 1,
            pattern of connections between inputs and outputs of conv layers at the two ends,
            should divide `groups`
        base_filter_length: int, default 1,
            lengths (sizes) of the filter kernels for conv layers at the two ends
        attn: dict, optional,
            attention mechanism for the neck conv layer,
            if None, no attention mechanism is used,
            keys:
                "name": str, can be "se", "gc", "nl" (alias "nonlocal", "non-local"), etc.
                "pos": int, position of the attention mechanism,
                other keys are specific to the attention mechanism
        config: dict,
            other hyper-parameters, including
            increase channel method, subsample method, dropout,
            activation choices, weight initializer, and short cut patterns, etc.

        """
        super().__init__()
        self.__num_convs = 3
        self.config = CFG(deepcopy(config))
        if self.__DEBUG__:
            print(
                f"configuration of {self.__name__} is as follows\n{dict_to_str(self.config)}"
            )
        self.expansion = self.config.get("expansion", self.expansion)

        self.__in_channels = in_channels
        # update denominator of computing neck_num_filters by init_num_filters
        self.__DEFAULT_BASE_WIDTH__ = self.config.get(
            "init_num_filters", self.__DEFAULT_BASE_WIDTH__
        )
        neck_num_filters = (
            int(num_filters * (base_width / self.__DEFAULT_BASE_WIDTH__)) * groups
        )
        self.__out_channels = [
            neck_num_filters,
            neck_num_filters,
            num_filters * self.expansion,
        ]
        if self.__DEBUG__:
            print(
                f"__DEFAULT_BASE_WIDTH__ = {self.__DEFAULT_BASE_WIDTH__}, "
                f"in_channels = {in_channels}, num_filters = {num_filters}, "
                f"base_width = {base_width}, neck_num_filters = {neck_num_filters}."
            )
        self.__base_filter_length = base_filter_length
        self.__kernel_size = [
            base_filter_length,
            filter_length,
            base_filter_length,
        ]
        self.__down_scale = subsample_length
        self.__stride = subsample_length
        if groups % base_groups != 0:
            raise ValueError("`groups` should be divisible by `base_groups`")
        self.__base_groups = base_groups
        self.__groups = [
            base_groups,
            groups,
            base_groups,
        ]
        self.__attn = attn
        if self.__attn:
            self.__attn = CFG(self.__attn)

        if (
            self.config.increase_channels_method.lower() == "zero_padding"
            and self.__groups != 1
        ):
            raise ValueError(
                "zero padding for increasing channels can not be used with groups != 1"
            )

        self.__increase_channels = self.__out_channels[-1] > self.__in_channels
        self.shortcut = self._make_shortcut_layer()

        self.main_stream = nn.Sequential()
        conv_names = {
            0: "cba_head",
            1: "cba_neck",
            2: "cba_tail",
        }
        conv_in_channels = self.__in_channels
        for i in range(self.__num_convs):
            if self.__attn and self.__attn["pos"] == i:
                self.main_stream.add_module(
                    self.__attn["name"],
                    make_attention_layer(conv_in_channels, **self.__attn),
                )
            conv_activation = (
                self.config.activation if i < self.__num_convs - 1 else None
            )
            conv_out_channels = self.__out_channels[i]
            conv_type = self.config.get("conv_type", None)
            if self.__kernel_size[i] == 1 and conv_type == "separable":
                conv_type = None
            self.main_stream.add_module(
                conv_names[i],
                Conv_Bn_Activation(
                    in_channels=conv_in_channels,
                    out_channels=conv_out_channels,
                    kernel_size=self.__kernel_size[i],
                    stride=(self.__stride if i == self.config.subsample_at else 1),
                    groups=self.__groups[i],
                    norm=True,
                    activation=conv_activation,
                    kw_activation=self.config.kw_activation,
                    kernel_initializer=self.config.kernel_initializer,
                    kw_initializer=self.config.kw_initializer,
                    bias=self.config.bias,
                    conv_type=conv_type,
                ),
            )
            if self.config.get("dropout", 0) > 0 and i < self.__num_convs - 1:
                self.main_stream.add_module(
                    f"dropout_{i}", nn.Dropout(self.config.dropout)
                )
            conv_in_channels = conv_out_channels
        if self.__attn and self.__attn["pos"] == -1:
            self.main_stream.add_module(
                self.__attn["name"],
                make_attention_layer(conv_in_channels, **self.__attn),
            )

        if isinstance(self.config.activation, str):
            self.out_activation = Activations[self.config.activation.lower()](
                **self.config.kw_activation
            )
        else:
            self.out_activation = self.config.activation(**self.config.kw_activation)

        if self.config.get("dropout", 0) > 0:
            self.out_dropout = nn.Dropout(self.config.dropout)
        else:
            self.out_dropout = None

    def _make_shortcut_layer(self) -> Union[nn.Module, None]:
        """ """
        if self.__DEBUG__:
            print(
                f"down_scale = {self.__down_scale}, increase_channels = {self.__increase_channels}"
            )
        if self.__down_scale > 1 or self.__increase_channels:
            if self.config.increase_channels_method.lower() == "conv":
                shortcut = DownSample(
                    down_scale=self.__down_scale,
                    in_channels=self.__in_channels,
                    out_channels=self.__out_channels[-1],
                    groups=self.__base_groups,
                    norm=True,
                    mode=self.config.subsample_mode,
                    filt_size=self.config.get("filt_size", 3),  # for blur pool
                )
            elif self.config.increase_channels_method.lower() == "zero_padding":
                batch_norm = (
                    False if self.config.subsample_mode.lower() != "conv" else True
                )
                shortcut = nn.Sequential(
                    DownSample(
                        down_scale=self.__down_scale,
                        in_channels=self.__in_channels,
                        out_channels=self.__in_channels,
                        norm=batch_norm,
                        mode=self.config.subsample_mode,
                        filt_size=self.config.get("filt_size", 3),  # for blur pool
                    ),
                    ZeroPadding(self.__in_channels, self.__out_channels[-1]),
                )
        else:
            shortcut = None
        return shortcut

    def forward(self, input: Tensor) -> Tensor:
        """

        Parameters
        ----------
        input: Tensor,
            of shape (batch_size, n_channels, seq_len)

        Returns
        -------
        out: Tensor,
            of shape (batch_size, n_channels, seq_len)

        """
        identity = input

        out = self.main_stream(input)

        if self.shortcut is not None:
            identity = self.shortcut(input)

        if identity.shape[-1] < out.shape[-1]:
            # downsampling using "avg" or "max" adopts no padding
            # hence shape[-1] of identity might be smaller by 1 than shape[-1] of out
            diff_sig_len = out.shape[-1] - identity.shape[-1]
            identity = F.pad(
                identity, [diff_sig_len // 2, diff_sig_len - diff_sig_len // 2]
            )

        out += identity
        out = self.out_activation(out)

        if self.out_dropout is not None:
            out = self.out_dropout(out)

        return out

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """

        Parameters
        ----------
        seq_len: int,
            length of the 1d sequence
        batch_size: int, optional,
            the batch size, can be None

        Returns
        -------
        output_shape: sequence,
            the output shape of this block, given `seq_len` and `batch_size`

        """
        _seq_len = seq_len
        for module in self.main_stream:
            if isinstance(module, nn.Dropout):
                continue
            output_shape = module.compute_output_shape(_seq_len, batch_size)
            _, _, _seq_len = output_shape
        return output_shape


class ResNetMacroBlock(nn.Sequential, SizeMixin):
    """NOT finished, NOT checked,"""

    __DEBUG__ = True
    __name__ = "ResNetMacroBlock"

    def __init__(self) -> NoReturn:
        """ """
        super().__init__()
        raise NotImplementedError

    def forward(self):
        """ """
        raise NotImplementedError

    def compute_output_shape(self):
        """ """
        raise NotImplementedError


class ResNetStem(nn.Sequential, SizeMixin):
    """
    the input stem of ResNet
    """

    __name__ = "ResNetStem"

    def __init__(
        self,
        in_channels: int,
        out_channels: Union[int, Sequence[int]],
        filter_lengths: Union[int, Sequence[int]],
        conv_stride: int,
        pool_size: int,
        pool_stride: int,
        subsample_mode: str = "max",
        groups: int = 1,
        **config,
    ) -> NoReturn:
        f"""

        Parameters
        ----------
        in_channels: int,
            the number of input channels
        out_channels: int or sequence of int,
            the number of output channels
        filter_lengths: int or sequence of int,
            the length of the filter, or equivalently,
            the kernel size(s) of the convolutions
        conv_stride: int,
            the stride of the convolution
        pool_size: int,
            the size of the pooling window
        pool_stride: int,
            the stride of the pooling window
        subsample_mode: str,
            the mode of subsampling, can be one of
            {DownSample.__MODES__},
            or "s2d" (with aliases "space_to_depth", "SpaceToDepth")
        groups: int,
            the number of groups for the convolution
        config: dict,
            the other configs for convolution and pooling

        """
        super().__init__()
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__filter_lengths = filter_lengths
        if subsample_mode.lower() in ["s2d", "space_to_depth", "SpaceToDepth"]:
            self.add_module(
                "s2d",
                SpaceToDepth(
                    self.__in_channels, self.__out_channels, config.get("block_size", 4)
                ),
            )
            return
        if isinstance(self.__filter_lengths, int):
            self.__filter_lengths = [self.__filter_lengths]
        if isinstance(self.__out_channels, int):
            self.__out_channels = [self.__out_channels]
        assert len(self.__filter_lengths) == len(self.__out_channels)

        conv_in_channels = self.__in_channels
        for idx, fl in enumerate(self.__filter_lengths):
            self.add_module(
                f"conv_{idx}",
                Conv_Bn_Activation(
                    conv_in_channels,
                    self.__out_channels[idx],
                    self.__filter_lengths[idx],
                    stride=conv_stride if idx == 0 else 1,
                    groups=groups,
                    **config,
                ),
            )
            conv_in_channels = self.__out_channels[idx]
        if pool_stride > 1:
            self.add_module(
                "pool",
                DownSample(
                    pool_stride,
                    conv_in_channels,
                    kernel_size=pool_size,
                    groups=groups,
                    padding=(pool_stride - 1) // 2,
                    mode=subsample_mode.lower(),
                    **config,
                ),
            )

    @add_docstring(compute_sequential_output_shape_docstring)
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """ """
        return compute_sequential_output_shape(self, seq_len, batch_size)


class ResNet(nn.Sequential, SizeMixin):
    """

    References
    ----------
    [1] https://github.com/awni/ecg
    [2] https://github.com/pytorch/vision/blob/master/torchvision/models/resnet.py

    TODO
    ----
    1. check performances of activations other than "nn.ReLU", especially mish and swish
    2. add functionality of "replace_stride_with_dilation"

    """

    __DEBUG__ = False
    __name__ = "ResNet"
    building_block = ResNetBasicBlock
    __DEFAULT_CONFIG__ = CFG(
        activation="relu",
        kw_activation={"inplace": True},
        kernel_initializer="he_normal",
        kw_initializer={},
        dropouts=0,
    )

    def __init__(self, in_channels: int, **config) -> NoReturn:
        """

        Parameters
        ----------
        in_channels: int,
            number of channels in the input
        config: dict,
            other hyper-parameters of the Module, ref. corresponding config file
            keyword arguments that have to be set:
            bias: bool,
                if True, each convolution will have a bias term
            num_blocks: sequence of int,
                number of building blocks in each macro block
            filter_lengths: int or sequence of int or sequence of sequences of int,
                filter length(s) (kernel size(s)) of the convolutions,
                with granularity to the whole network, to each macro block,
                or to each building block
            subsample_lengths: int or sequence of int or sequence of sequences of int,
                subsampling length(s) (ratio(s)) of all blocks,
                with granularity to the whole network, to each macro block,
                or to each building block,
                the former 2 subsample at the first building block
            groups: int,
                connection pattern (of channels) of the inputs and outputs
            stem: dict,
                other parameters that can be set for the input stem
            block: dict,
                other parameters that can be set for the building blocks
            for a full list of configurable parameters, ref. corr. config file

        """
        super().__init__()
        self.__in_channels = in_channels
        self.config = CFG(deepcopy(self.__DEFAULT_CONFIG__))
        self.config.update(deepcopy(config))
        if self.__DEBUG__:
            print(
                f"configuration of {self.__name__} is as follows\n{dict_to_str(self.config)}"
            )
        if not isinstance(self.config.get("building_block", None), str):
            # for those architectures that have multiple type of building blocks
            self.building_block = []
            self.additional_kw = []
            for b in self.config.building_block:
                if b.lower() in [
                    "bottleneck",
                    "bottle_neck",
                ]:
                    self.building_block.append(ResNetBottleNeck)
                    self.additional_kw.append(
                        CFG(
                            {
                                k: self.config[k]
                                for k in [
                                    "base_width",
                                    "base_groups",
                                    "base_filter_length",
                                ]
                                if k in self.config.keys()
                            }
                        )
                    )
                else:
                    self.building_block.append(ResNetBasicBlock)
                    self.additional_kw.append(CFG())
            assert isinstance(self.config.block, Sequence) and len(
                self.config.block
            ) == len(self.config.building_block) == len(self.config.num_blocks)
        elif self.config.get("building_block", "").lower() in [
            "bottleneck",
            "bottle_neck",
        ]:
            self.building_block = ResNetBottleNeck
            # additional parameters for bottleneck
            self.additional_kw = CFG(
                {
                    k: self.config[k]
                    for k in [
                        "base_width",
                        "base_groups",
                        "base_filter_length",
                    ]
                    if k in self.config.keys()
                }
            )
        else:
            self.additional_kw = CFG()
        if self.__DEBUG__:
            print(f"additional_kw = {self.additional_kw}")

        self.add_module(
            "input_stem",
            ResNetStem(
                in_channels=self.__in_channels,
                out_channels=self.config.stem.num_filters,
                # bottleneck use "base_groups"
                groups=self.config.get("base_groups", self.config.groups),
                activation=self.config.activation,
                **self.config.stem,
            ),
        )

        if isinstance(self.config.filter_lengths, int):
            self.__filter_lengths = list(
                repeat(self.config.filter_lengths, len(self.config.num_blocks))
            )
        else:
            self.__filter_lengths = self.config.filter_lengths
        assert len(self.__filter_lengths) == len(self.config.num_blocks), (
            f"`config.filter_lengths` indicates {len(self.__filter_lengths)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )
        if isinstance(self.config.subsample_lengths, int):
            self.__subsample_lengths = list(
                repeat(self.config.subsample_lengths, len(self.config.num_blocks))
            )
        else:
            self.__subsample_lengths = self.config.subsample_lengths
        assert len(self.__subsample_lengths) == len(self.config.num_blocks), (
            f"`config.subsample_lengths` indicates {len(self.__subsample_lengths)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )
        self.__num_filters = self.config.get(
            "num_filters",
            [
                (2**i) * self.input_stem.compute_output_shape()[1]
                for i in range(len(self.config.num_blocks))
            ],
        )
        assert len(self.__num_filters) == len(self.config.num_blocks), (
            f"`config.num_filters` indicates {len(self.__num_filters)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )
        if isinstance(self.config.dropouts, Real):
            self.__dropouts = list(
                repeat(self.config.dropouts, len(self.config.num_blocks))
            )
        else:
            self.__dropouts = self.config.dropouts
        assert len(self.__dropouts) == len(self.config.num_blocks), (
            f"`config.dropouts` indicates {len(self.__dropouts)} macro blocks, "
            f"while `config.num_blocks` indicates {len(self.config.num_blocks)}"
        )

        # grouped resnet (basic) blocks,
        # number of channels are doubled at the first block of each macro-block
        macro_in_channels = self.input_stem.compute_output_shape()[1]
        for macro_idx, nb in enumerate(self.config.num_blocks):
            macro_num_filters = self.__num_filters[macro_idx]
            macro_filter_lengths = self.__filter_lengths[macro_idx]
            macro_subsample_lengths = self.__subsample_lengths[macro_idx]
            block_in_channels = macro_in_channels
            block_num_filters = macro_num_filters
            if isinstance(macro_filter_lengths, int):
                block_filter_lengths = list(repeat(macro_filter_lengths, nb))
            else:
                block_filter_lengths = macro_filter_lengths
            assert len(block_filter_lengths) == nb, (
                f"at the {macro_idx}-th macro block, `macro_subsample_lengths` indicates {len(macro_subsample_lengths)} "
                f"building blocks, while `config.num_blocks[{macro_idx}]` indicates {nb}"
            )
            if isinstance(macro_subsample_lengths, int):
                # subsample at the first building block
                block_subsample_lengths = list(repeat(1, nb))
                block_subsample_lengths[0] = macro_subsample_lengths
            else:
                block_subsample_lengths = macro_subsample_lengths
            assert len(block_subsample_lengths) == nb, (
                f"at the {macro_idx}-th macro block, `macro_subsample_lengths` indicates {len(macro_subsample_lengths)} "
                f"building blocks, while `config.num_blocks[{macro_idx}]` indicates {nb}"
            )
            if isinstance(self.building_block, Sequence):
                bb = self.building_block[macro_idx]
                bb_config = self.config.block[macro_idx]
                bb_kw = self.additional_kw[macro_idx]
            else:
                bb = self.building_block
                bb_config = self.config.block
                bb_kw = self.additional_kw
            for block_idx in range(nb):
                self.add_module(
                    f"{bb.__name__}_{macro_idx}_{block_idx}",
                    bb(
                        in_channels=block_in_channels,
                        num_filters=block_num_filters,
                        filter_length=block_filter_lengths[block_idx],
                        subsample_length=block_subsample_lengths[block_idx],
                        groups=self.config.groups,
                        dilation=1,
                        dropout=self.__dropouts[macro_idx],
                        **(bb_kw),
                        **(bb_config),
                    ),
                )
                block_in_channels = block_num_filters * bb.expansion
            macro_in_channels = macro_num_filters * bb.expansion

    @add_docstring(compute_sequential_output_shape_docstring)
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """ """
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels



# -*- coding: utf-8 -*-
"""
docstring, to write
"""

from .cpsc2018 import CPSC2018
from .cpsc2019 import CPSC2019
from .cpsc2020 import CPSC2020
from .cpsc2021 import CPSC2021

__all__ = [
    "CPSC2018",
    "CPSC2019",
    "CPSC2020",
    "CPSC2021",
]



"""
"""

from ..cfg import CFG

__all__ = [
    "RR_AF_VANILLA_CONFIG",
    "RR_AF_CRF_CONFIG",
    "RR_LSTM_CONFIG",
]


RR_AF_VANILLA_CONFIG = CFG()

RR_AF_VANILLA_CONFIG.lstm = CFG()
RR_AF_VANILLA_CONFIG.lstm.bias = True
RR_AF_VANILLA_CONFIG.lstm.dropouts = 0.1
RR_AF_VANILLA_CONFIG.lstm.bidirectional = True
RR_AF_VANILLA_CONFIG.lstm.retseq = True
RR_AF_VANILLA_CONFIG.lstm.hidden_sizes = [200]

RR_AF_VANILLA_CONFIG.attn = CFG()
RR_AF_VANILLA_CONFIG.attn.name = "none"  # "gc", "nl", "se"

RR_AF_VANILLA_CONFIG.global_pool = "max"  # "avg", "attn"
RR_AF_VANILLA_CONFIG.global_pool_size = 1

RR_AF_VANILLA_CONFIG.clf = CFG()
RR_AF_VANILLA_CONFIG.clf.name = "linear"  # crf
RR_AF_VANILLA_CONFIG.clf.linear = CFG()
RR_AF_VANILLA_CONFIG.clf.linear.out_channels = [
    50,  # not including the last linear layer, with out channels equals n_classes
]
RR_AF_VANILLA_CONFIG.clf.linear.bias = True
RR_AF_VANILLA_CONFIG.clf.linear.dropouts = 0.1
RR_AF_VANILLA_CONFIG.clf.linear.activation = "relu"


RR_AF_CRF_CONFIG = CFG()

RR_AF_CRF_CONFIG.lstm = CFG()
RR_AF_CRF_CONFIG.lstm.bias = True
RR_AF_CRF_CONFIG.lstm.dropouts = 0.1
RR_AF_CRF_CONFIG.lstm.bidirectional = True
RR_AF_CRF_CONFIG.lstm.retseq = True
RR_AF_CRF_CONFIG.lstm.hidden_sizes = [200]

RR_AF_CRF_CONFIG.attn = CFG()
RR_AF_CRF_CONFIG.attn.name = "none"  # "gc", "nl", "se"

RR_AF_CRF_CONFIG.clf = CFG()
RR_AF_CRF_CONFIG.clf.name = "crf"  # crf
RR_AF_CRF_CONFIG.clf.crf = CFG()
RR_AF_CRF_CONFIG.clf.crf.proj_bias = True


RR_LSTM_CONFIG = CFG()

RR_LSTM_CONFIG.lstm = CFG()
RR_LSTM_CONFIG.lstm.bias = True
RR_LSTM_CONFIG.lstm.dropouts = 0.1
RR_LSTM_CONFIG.lstm.bidirectional = True
RR_LSTM_CONFIG.lstm.retseq = True
RR_LSTM_CONFIG.lstm.hidden_sizes = [200]

RR_LSTM_CONFIG.attn = CFG()
RR_LSTM_CONFIG.attn.name = "se"  # "gc", "nl", "none"
RR_LSTM_CONFIG.attn.se = CFG()
RR_LSTM_CONFIG.attn.se.reduction = 8  # not including the last linear layer
RR_LSTM_CONFIG.attn.se.activation = "relu"
RR_LSTM_CONFIG.attn.se.kw_activation = CFG(inplace=True)
RR_LSTM_CONFIG.attn.se.bias = True
RR_LSTM_CONFIG.attn.se.kernel_initializer = "he_normal"


RR_LSTM_CONFIG.global_pool = "avg"  # "avg", "attn", "none"


RR_LSTM_CONFIG.clf = CFG()
RR_LSTM_CONFIG.clf.name = "crf"  # "linear"

RR_LSTM_CONFIG.clf.crf = CFG()
RR_LSTM_CONFIG.clf.crf.proj_bias = True

# RR_LSTM_CONFIG.clf.linear = CFG()



"""
"""

from typing import Any, NoReturn, Optional, Sequence, Union

import torch
import torch.nn as nn

from ...cfg import DEFAULTS
from ...utils.utils_nn import SizeMixin

if DEFAULTS.torch_dtype == torch.float64:
    torch.set_default_tensor_type(torch.DoubleTensor)


__all__ = [
    "Transformer",
]


class Transformer(nn.Module, SizeMixin):
    """ """

    __DEBUG__ = True
    __name__ = "Transformer"

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        num_heads: int = 8,
        num_layers: int = 1,
        dropout: float = 0.1,
        **kwargs: Any,
    ) -> NoReturn:
        """

        Parameters
        ----------
        input_size: int,
            number of input channels
        hidden_size: int,
            number of hidden units in the encoding layer
        num_heads: int, default 8,
            number of attention heads
        num_layers: int, default 1,
            number of encoding layers
        dropout: float, default 0.1,
            dropout probability
        kwargs: keyword arguments,

        """
        super().__init__()
        self.__input_size = input_size
        self.__hidden_size = hidden_size
        self.__num_layers = num_layers
        self.__num_heads = num_heads
        self.__batch_first = kwargs.get("batch_first", False)
        try:
            encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.__input_size,
                nhead=self.__num_heads,
                dim_feedforward=self.__hidden_size,
                dropout=dropout,
                batch_first=self.__batch_first,
                activation=kwargs.get("activation", "relu"),
            )
        except Exception:
            self.__batch_first = False
            encoder_layer = nn.TransformerEncoderLayer(
                d_model=self.__input_size,
                nhead=self.__num_heads,
                dim_feedforward=self.__hidden_size,
                dropout=dropout,
                activation=kwargs.get("activation", "relu"),
            )
            print("batch_first only supports torch >= 1.9, defaults to False")
        self.encoder = nn.TransformerEncoder(
            encoder_layer, num_layers=self.__num_layers
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """

        Parameters
        ----------
        x: torch.Tensor,
            the input tensor,
            of shape (seq_len, batch_size, input_size) or (batch_size, seq_len, input_size)

        Returns
        -------
        torch.Tensor:
            of shape (seq_len, batch_size, input_size) or (batch_size, seq_len, input_size)

        """
        return self.encoder(x)

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """

        Parameters
        ----------
        seq_len: int,
            length of the 1d sequence
        batch_size: int, optional,
            the batch size, can be None

        Returns
        -------
        output_shape: sequence,
            the output shape of this layer, given `seq_len` and `batch_size`

        """
        if self.__batch_first:
            return (batch_size, seq_len, self.__input_size)
        else:
            return (seq_len, batch_size, self.__input_size)



# -*- coding: utf-8 -*-
"""
"""

import math
from numbers import Real
from pathlib import Path
from typing import Any, Dict, List, NoReturn, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from scipy.io import loadmat

from ...cfg import CFG, DEFAULTS
from ...utils.misc import add_docstring
from ...utils.download import http_get
from ...utils.ecg_arrhythmia_knowledge import PVC, SPB  # noqa: F401
from ...utils.utils_interval import get_optimal_covering
from ..base import DEFAULT_FIG_SIZE_PER_SEC, CPSCDataBase, DataBaseInfo


__all__ = [
    "CPSC2020",
    "compute_metrics",
]


_CPSC2020_INFO = DataBaseInfo(
    title="""
    The 3rd China Physiological Signal Challenge 2020:
    Searching for Premature Ventricular Contraction (PVC) and Supraventricular Premature Beat (SPB) from Long-term ECGs
    """,
    about="""
    1. training data consists of 10 single-lead ECG recordings collected from arrhythmia patients, each of the recording last for about 24 hours
    2. data and annotations are stored in v5 .mat files
    3. A02, A03, A08 are patient with atrial fibrillation
    4. sampling frequency = 400 Hz
    5. Detailed information:
        rec   ?AF   Length(h)   # N beats   # V beats   # S beats   # Total beats
        A01   No	25.89       109,062     0           24          109,086
        A02   Yes	22.83       98,936      4,554       0           103,490
        A03   Yes	24.70       137,249     382         0           137,631
        A04   No	24.51       77,812      19,024      3,466       100,302
        A05   No	23.57       94,614  	1	        25	        94,640
        A06   No	24.59       77,621  	0	        6	        77,627
        A07   No	23.11	    73,325  	15,150	    3,481	    91,956
        A08   Yes	25.46	    115,518 	2,793	    0	        118,311
        A09   No	25.84	    88,229  	2	        1,462	    89,693
        A10   No	23.64	    72,821	    169	        9,071	    82,061
    6. challenging factors for accurate detection of SPB and PVC:
        amplitude variation; morphological variation; noise
    """,
    note="""
    1. the records can roughly be classified into 4 groups:
        N:  A01, A03, A05, A06
        V:  A02, A08
        S:  A09, A10
        VS: A04, A07
    2. as premature beats and atrial fibrillation can co-exists
        (via the following code, and data from CINC2020),
        the situation becomes more complicated.
        >>> from utils.scoring_aux_data import dx_cooccurrence_all
        >>> dx_cooccurrence_all.loc["AF", ["PAC","PVC","SVPB","VPB"]]
        PAC     20
        PVC     19
        SVPB     4
        VPB     20
        Name: AF, dtype: int64
        this could also be seen from this dataset, via the following code as an example:
        >>> from data_reader import CPSC2020Reader as CR
        >>> db_dir = "/media/cfs/wenhao71/data/CPSC2020/TrainingSet/"
        >>> dr = CR(db_dir)
        >>> rec = dr.all_records[1]
        >>> dr.plot(rec, sampfrom=0, sampto=4000, ticks_granularity=2)
    3. PVC and SPB can also co-exist, as illustrated via the following code (from CINC2020):
        >>> from utils.scoring_aux_data import dx_cooccurrence_all
        >>> dx_cooccurrence_all.loc[["PVC","VPB"], ["PAC","SVPB",]]
        PAC	SVPB
        PVC	14	1
        VPB	27	0
        and also from the following code:
        >>> for rec in dr.all_records:
        >>>     ann = dr.load_ann(rec)
        >>>     spb = ann["SPB_indices"]
        >>>     pvc = ann["PVC_indices"]
        >>>     if len(np.diff(spb)) > 0:
        >>>         print(f"{rec}: min dist among SPB = {np.min(np.diff(spb))}")
        >>>     if len(np.diff(pvc)) > 0:
        >>>         print(f"{rec}: min dist among PVC = {np.min(np.diff(pvc))}")
        >>>     diff = [s-p for s,p in product(spb, pvc)]
        >>>     if len(diff) > 0:
        >>>         print(f"{rec}: min dist between SPB and PVC = {np.min(np.abs(diff))}")
        A01: min dist among SPB = 630
        A02: min dist among SPB = 696
        A02: min dist among PVC = 87
        A02: min dist between SPB and PVC = 562
        A03: min dist among SPB = 7044
        A03: min dist among PVC = 151
        A03: min dist between SPB and PVC = 3750
        A04: min dist among SPB = 175
        A04: min dist among PVC = 156
        A04: min dist between SPB and PVC = 178
        A05: min dist among SPB = 182
        A05: min dist between SPB and PVC = 22320
        A06: min dist among SPB = 455158
        A07: min dist among SPB = 603
        A07: min dist among PVC = 153
        A07: min dist between SPB and PVC = 257
        A08: min dist among SPB = 2903029
        A08: min dist among PVC = 106
        A08: min dist between SPB and PVC = 350
        A09: min dist among SPB = 180
        A09: min dist among PVC = 7719290
        A09: min dist between SPB and PVC = 1271
        A10: min dist among SPB = 148
        A10: min dist among PVC = 708
        A10: min dist between SPB and PVC = 177
    """,
    usage=[
        "ECG arrhythmia (PVC, SPB) detection",
    ],
    issues="""
    1. currently, using `xqrs` as qrs detector,
       a lot more (more than 1000) rpeaks would be detected for A02, A07, A08,
       which might be caused by motion artefacts (or AF?);
       a lot less (more than 1000) rpeaks would be detected for A04.
       numeric details are as follows:
       ----------------------------------------------
       rec   ?AF    # beats by xqrs     # Total beats
       A01   No     109502              109,086
       A02   Yes    119562              103,490
       A03   Yes    135912              137,631
       A04   No     92746               100,302
       A05   No     94674               94,640
       A06   No     77955               77,627
       A07   No     98390               91,956
       A08   Yes    126908              118,311
       A09   No     89972               89,693
       A10   No     83509               82,061
    2. (fixed by an official update) A04 has duplicate "PVC_indices" (13534856,27147621,35141190 all appear twice):
       before correction of `load_ann`:
       >>> from collections import Counter
       >>> db_dir = "/mnt/wenhao71/data/CPSC2020/TrainingSet/"
       >>> data_gen = CPSC2020Reader(db_dir=db_dir,working_dir=db_dir)
       >>> rec = 4
       >>> ann = data_gen.load_ann(rec)
       >>> Counter(ann["PVC_indices"]).most_common()[:4]
       would produce [(13534856, 2), (27147621, 2), (35141190, 2), (848, 1)]
    3. when extracting morphological features using augmented rpeaks for A04,
       `RuntimeWarning: invalid value encountered in double_scalars` would raise
       for `R_value = (R_value - y_min) / (y_max - y_min)` and
       for `y_values[n] = (y_values[n] - y_min) / (y_max - y_min)`.
       this is caused by the 13882273-th sample, which is contained in "PVC_indices",
       however, whether it is a PVC beat, or just motion artefact, is in doubt!
    """,
    references=[
        "http://www.icbeb.org/CPSC2020.html",
        "https://github.com/PIA-Group/BioSPPy",
    ],
    doi="10.1166/jmihi.2020.3289",
)


@add_docstring(_CPSC2020_INFO.format_database_docstring())
class CPSC2020(CPSCDataBase):
    """ """

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 2,
        **kwargs: Any,
    ) -> NoReturn:
        """

        Parameters
        ----------
        db_dir: str or Path, optional,
            storage path of the database
        working_dir: str or Path, optional,
            working directory, to store intermediate files and log file
        verbose: int, default 2,
            log verbosity
        kwargs: auxilliary key word arguments
        """
        super().__init__(
            db_name="cpsc2020",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )

        self.fs = 400
        self.spacing = 1000 / self.fs
        self.rec_ext = "mat"
        self.ann_ext = "mat"

        self.rec_dir = self.db_dir / "data"
        self.ann_dir = self.db_dir / "ref"
        # aliases
        self.data_dir = self.rec_dir
        self.ref_dir = self.ann_dir

        self.n_records = 10
        self._all_records = None
        self._all_annotations = None
        self._ls_rec()

        self.subgroups = CFG(
            {
                "N": [
                    "A01",
                    "A03",
                    "A05",
                    "A06",
                ],
                "V": ["A02", "A08"],
                "S": ["A09", "A10"],
                "VS": ["A04", "A07"],
            }
        )

        self.palette = {
            "spb": "yellow",
            "pvc": "red",
        }

    def _ls_rec(self) -> NoReturn:
        """ """
        self._all_records = [f"A{i:02d}" for i in range(1, 1 + self.n_records)]
        self._all_annotations = [f"R{i:02d}" for i in range(1, 1 + self.n_records)]
        self._df_records = pd.DataFrame()
        self._df_records["record"] = self._all_records
        self._df_records["path"] = self._df_records["record"].apply(
            lambda x: self.get_absolute_path(x)
        )
        self._df_records["annotation"] = self._all_annotations
        self._df_records.set_index("record", inplace=True)

    @property
    def all_annotations(self):
        """ """
        return self._all_annotations

    @property
    def all_references(self):
        """ """
        return self._all_annotations

    def get_subject_id(self, rec: Union[int, str]) -> int:
        """not finished,

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`

        Returns
        -------
        pid: int,
            the `subject_id` corr. to `rec_no`
        """
        pid = 0
        raise NotImplementedError

    def get_absolute_path(
        self,
        rec: Union[str, int],
        extension: Optional[str] = None,
        ann: bool = False,
    ) -> Path:
        """
        get the absolute path of the record `rec`

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        extension: str, optional,
            extension of the file
        ann: bool, default False,
            whether to get the annotation file path or not

        Returns
        -------
        Path,
            absolute path of the file

        """
        if isinstance(rec, int):
            rec = self[rec]
        if extension is not None and not extension.startswith("."):
            extension = f".{extension}"
        if ann:
            rec = rec.replace("A", "R")
            return self.ann_dir / f"{rec}{extension or ''}"
        return self.data_dir / f"{rec}{extension or ''}"

    def load_data(
        self,
        rec: Union[int, str],
        units: str = "mV",
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        keep_dim: bool = True,
    ) -> np.ndarray:
        """

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        units: str, default "mV",
            units of the output signal, can also be "μV", with an alias of "uV"
        sampfrom: int, optional,
            start index of the data to be loaded
        sampto: int, optional,
            end index of the data to be loaded
        keep_dim: bool, default True,
            whether or not to flatten the data of shape (n,1)

        Returns
        -------
        data: ndarray,
            the ECG data

        """
        rec_fp = self.get_absolute_path(rec, self.rec_ext)
        data = loadmat(str(rec_fp))["ecg"]
        if units.lower() in ["uv", "μv"]:
            data = (1000 * data).astype(int)
        sf, st = (sampfrom or 0), (sampto or len(data))
        data = data[sf:st]
        if not keep_dim:
            data = data.flatten()
        return data

    def load_ann(
        self,
        rec: Union[int, str],
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
    ) -> Dict[str, np.ndarray]:
        """

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        sampfrom: int, optional,
            start index of the data to be loaded
        sampto: int, optional,
            end index of the data to be loaded

        Returns
        -------
        ann: dict,
            with items (ndarray) "SPB_indices" and "PVC_indices",
            which record the indices of SPBs and PVCs
        """
        ann_fp = self.get_absolute_path(rec, self.ann_ext, ann=True)
        ann = loadmat(str(ann_fp))["ref"]
        sf, st = (sampfrom or 0), (sampto or np.inf)
        spb_indices = ann["S_ref"][0, 0].flatten().astype(int)
        # drop duplicates
        spb_indices = np.array(sorted(list(set(spb_indices))), dtype=int)
        spb_indices = spb_indices[np.where((spb_indices >= sf) & (spb_indices < st))[0]]
        pvc_indices = ann["V_ref"][0, 0].flatten().astype(int)
        # drop duplicates
        pvc_indices = np.array(sorted(list(set(pvc_indices))), dtype=int)
        pvc_indices = pvc_indices[np.where((pvc_indices >= sf) & (pvc_indices < st))[0]]
        ann = {
            "SPB_indices": spb_indices,
            "PVC_indices": pvc_indices,
        }
        return ann

    def train_test_split_rec(self, test_rec_num: int = 2) -> Dict[str, List[str]]:
        """

        split the records into train set and test set

        Parameters
        ----------
        test_rec_num: int,
            number of records for the test set

        Returns
        -------
        split_res: dict,
            with items `train`, `test`, both being list of record names
        """
        if test_rec_num == 1:
            test_records = DEFAULTS.RNG_sample(self.subgroups.VS, 1).tolist()
        elif test_rec_num == 2:
            test_records = (
                DEFAULTS.RNG_sample(self.subgroups.VS, 1).tolist()
                + DEFAULTS.RNG_sample(self.subgroups.N, 1).tolist()
            )
        elif test_rec_num == 3:
            test_records = (
                DEFAULTS.RNG_sample(self.subgroups.VS, 1).tolist()
                + DEFAULTS.RNG_sample(self.subgroups.N, 2).tolist()
            )
        elif test_rec_num == 4:
            test_records = []
            for k in self.subgroups.keys():
                test_records += DEFAULTS.RNG_sample(self.subgroups[k], 1).tolist()
        else:
            raise ValueError("test data ratio too high")
        train_records = [r for r in self.all_records if r not in test_records]

        split_res = CFG(
            {
                "train": train_records,
                "test": test_records,
            }
        )

        return split_res

    def locate_premature_beats(
        self,
        rec: Union[int, str],
        premature_type: Optional[str] = None,
        window: int = 10000,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
    ) -> List[List[int]]:
        """

        locate the sample indices of premature beats in a record,
        in the form of a list of lists,
        each list contains the interval of sample indices of premature beats

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        premature_type: str, optional,
            premature beat type, can be one of "SPB", "PVC"
        window: int, default 10000,
            window length of each premature beat
        sampfrom: int, optional,
            start index of the premature beats to locate
        sampto: int, optional,
            end index of the premature beats to locate

        Returns
        -------
        premature_intervals: list,
            list of intervals of premature beats

        """
        ann = self.load_ann(rec)
        if premature_type:
            premature_inds = ann[f"{premature_type.upper()}_indices"]
        else:
            premature_inds = np.append(ann["SPB_indices"], ann["PVC_indices"])
            premature_inds = np.sort(premature_inds)
        try:  # premature_inds empty?
            sf, st = (sampfrom or 0), (sampto or premature_inds[-1] + 1)
        except Exception:
            premature_intervals = []
            return premature_intervals
        premature_inds = premature_inds[(sf < premature_inds) & (premature_inds < st)]
        tot_interval = [sf, st]
        premature_intervals, _ = get_optimal_covering(
            total_interval=tot_interval,
            to_cover=premature_inds,
            min_len=window * self.fs // 1000,
            split_threshold=window * self.fs // 1000,
            traceback=False,
        )
        return premature_intervals

    def plot(
        self,
        rec: Union[int, str],
        data: Optional[np.ndarray] = None,
        ann: Optional[Dict[str, np.ndarray]] = None,
        ticks_granularity: int = 0,
        sampfrom: Optional[int] = None,
        sampto: Optional[int] = None,
        rpeak_inds: Optional[Union[Sequence[int], np.ndarray]] = None,
    ) -> NoReturn:
        """

        Parameters
        ----------
        rec: str or int,
            record name or index of the record in `self.all_records`
        data: ndarray, optional,
            ECG signal to plot,
            if given, data of `rec` will not be used,
            this is useful when plotting filtered data
        ann: dict, optional,
            annotations for `data`, covering those from annotation files,
            "SPB_indices", "PVC_indices", each of ndarray values,
            ignored if `data` is None
        ticks_granularity: int, default 0,
            the granularity to plot axis ticks, the higher the more,
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        sampfrom: int, optional,
            start index of the data to plot
        sampto: int, optional,
            end index of the data to plot
        rpeak_inds: array_like, optional,
            indices of R peaks,
            if `data` is None, then indices should be the absolute indices in the record
        """
        if isinstance(rec, int):
            rec = self[rec]
        if "plt" not in dir():
            import matplotlib.pyplot as plt
        import matplotlib.patches as mpatches

        patches = {}

        if data is None:
            _data = self.load_data(
                rec, units="μV", sampfrom=sampfrom, sampto=sampto, keep_dim=False
            )
        else:
            units = self._auto_infer_units(data)
            if units == "mV":
                _data = data * 1000
            elif units == "μV":
                _data = data.copy()

        if ann is None or data is None:
            ann = self.load_ann(rec, sampfrom=sampfrom, sampto=sampto)
        sf, st = (sampfrom or 0), (sampto or len(_data))
        spb_indices = ann["SPB_indices"]
        pvc_indices = ann["PVC_indices"]
        spb_indices = spb_indices - sf
        pvc_indices = pvc_indices - sf

        if rpeak_inds is not None:
            if data is not None:
                rpeak_secs = np.array(rpeak_inds) / self.fs
            else:
                rpeak_secs = np.array(rpeak_inds)
                rpeak_secs = rpeak_secs[
                    np.where((rpeak_secs >= sf) & (rpeak_secs < st))[0]
                ]
                rpeak_secs = (rpeak_secs - sf) / self.fs

        line_len = self.fs * 25  # 25 seconds
        nb_lines = math.ceil(len(_data) / line_len)

        bias_thr = 0.15
        winL = 0.06
        winR = 0.08

        for idx in range(nb_lines):
            seg = _data[idx * line_len : (idx + 1) * line_len]
            secs = (np.arange(len(seg)) + idx * line_len) / self.fs
            fig_sz_w = int(round(DEFAULT_FIG_SIZE_PER_SEC * len(seg) / self.fs))
            y_range = np.max(np.abs(seg)) + 100
            fig_sz_h = 6 * y_range / 1500
            fig, ax = plt.subplots(figsize=(fig_sz_w, fig_sz_h))
            ax.plot(
                secs,
                seg,
                color="black",
                linewidth="2.0",
            )
            ax.axhline(y=0, linestyle="-", linewidth="1.0", color="red")
            if ticks_granularity >= 1:
                ax.xaxis.set_major_locator(plt.MultipleLocator(0.2))
                ax.yaxis.set_major_locator(plt.MultipleLocator(500))
                ax.grid(which="major", linestyle="-", linewidth="0.5", color="red")
            if ticks_granularity >= 2:
                ax.xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                ax.yaxis.set_minor_locator(plt.MultipleLocator(100))
                ax.grid(which="minor", linestyle=":", linewidth="0.5", color="black")
            seg_spb = np.where(
                (spb_indices >= idx * line_len) & (spb_indices < (idx + 1) * line_len)
            )[0]
            # print(f"spb_indices = {spb_indices}, seg_spb = {seg_spb}")
            if len(seg_spb) > 0:
                seg_spb = spb_indices[seg_spb] / self.fs
                patches["SPB"] = mpatches.Patch(color=self.palette["spb"], label="SPB")
            seg_pvc = np.where(
                (pvc_indices >= idx * line_len) & (pvc_indices < (idx + 1) * line_len)
            )[0]
            # print(f"pvc_indices = {pvc_indices}, seg_pvc = {seg_pvc}")
            if len(seg_pvc) > 0:
                seg_pvc = pvc_indices[seg_pvc] / self.fs
                patches["PVC"] = mpatches.Patch(color=self.palette["pvc"], label="PVC")
            for t in seg_spb:
                ax.axvspan(
                    max(secs[0], t - bias_thr),
                    min(secs[-1], t + bias_thr),
                    color=self.palette["spb"],
                    alpha=0.3,
                )
                ax.axvspan(
                    max(secs[0], t - winL),
                    min(secs[-1], t + winR),
                    color=self.palette["spb"],
                    alpha=0.9,
                )
            for t in seg_pvc:
                ax.axvspan(
                    max(secs[0], t - bias_thr),
                    min(secs[-1], t + bias_thr),
                    color=self.palette["pvc"],
                    alpha=0.3,
                )
                ax.axvspan(
                    max(secs[0], t - winL),
                    min(secs[-1], t + winR),
                    color=self.palette["pvc"],
                    alpha=0.9,
                )
            if len(patches) > 0:
                ax.legend(
                    handles=[v for _, v in patches.items()],
                    loc="lower left",
                    prop={"size": 16},
                )
            if rpeak_inds is not None:
                seg_rpeak_secs = rpeak_secs[
                    np.where((rpeak_secs >= secs[0]) & (rpeak_secs < secs[-1]))[0]
                ]
                for r in seg_rpeak_secs:
                    ax.axvspan(r - 0.01, r + 0.01, color="green", alpha=0.7)
            ax.set_xlim(secs[0], secs[-1])
            ax.set_ylim(-y_range, y_range)
            ax.set_xlabel("Time [s]")
            ax.set_ylabel("Voltage [μV]")
            plt.show()


def _ann_to_beat_ann_epoch_v1(
    rpeaks: np.ndarray, ann: Dict[str, np.ndarray], bias_thr: Real
) -> dict:
    """

    the naive method to label beat types using annotations provided by the dataset

    Parameters
    ----------
    rpeaks: ndarray,
        rpeaks for forming beats
    ann: dict,
        with items (ndarray) "SPB_indices" and "PVC_indices",
        which record the indices of SPBs and PVCs
    bias_thr: real number,
        tolerance for using annotations (PVC, SPB indices provided by the dataset),
        to label the type of beats given by `rpeaks`

    Returns
    -------
    retval: dict, with the following items
        - ann_matched: dict of ndarray,
            indices of annotations ("SPB_indices" and "PVC_indices")
            that match some beat from `rpeaks`.
            for v1, this term is always the same as `ann`, hence useless
        - beat_ann: ndarray,
            label for each beat from `rpeaks`

    """
    beat_ann = np.array(["N" for _ in range(len(rpeaks))])
    for idx, r in enumerate(rpeaks):
        if any([abs(r - p) < bias_thr for p in ann["SPB_indices"]]):
            beat_ann[idx] = "S"
        elif any([abs(r - p) < bias_thr for p in ann["PVC_indices"]]):
            beat_ann[idx] = "V"
    ann_matched = ann.copy()
    retval = dict(ann_matched=ann_matched, beat_ann=beat_ann)
    return retval


def _ann_to_beat_ann_epoch_v2(
    rpeaks: np.ndarray, ann: Dict[str, np.ndarray], bias_thr: Real
) -> dict:
    """has flaws, deprecated,

    similar to `_ann_to_beat_ann_epoch_v1`, but records those matched annotations,
    for further post-process, adding those beats that are in annotation,
    but not detected by the signal preprocessing algorithms (qrs detection)

    however, the comparison process (the block inside the outer `for` loop)
    is not quite correct

    Parameters
    ----------
    rpeaks: ndarray,
        rpeaks for forming beats
    ann: dict,
        with items (ndarray) "SPB_indices" and "PVC_indices",
        which record the indices of SPBs and PVCs
    bias_thr: real number,
        tolerance for using annotations (PVC, SPB indices provided by the dataset),
        to label the type of beats given by `rpeaks`

    Returns
    -------
    retval: dict, with the following items
        - ann_matched: dict of ndarray,
            indices of annotations ("SPB_indices" and "PVC_indices")
            that match some beat from `rpeaks`
        - beat_ann: ndarray,
            label for each beat from `rpeaks`

    """
    beat_ann = np.array(["N" for _ in range(len(rpeaks))], dtype="<U1")
    # used to add back those beat that is not detected via proprocess algorithm
    _ann = {k: v.astype(int).tolist() for k, v in ann.items()}
    for idx_r, r in enumerate(rpeaks):
        found = False
        for idx_a, a in enumerate(_ann["SPB_indices"]):
            if abs(r - a) < bias_thr:
                found = True
                beat_ann[idx_r] = "S"
                del _ann["SPB_indices"][idx_a]
                break
        if found:
            continue
        for idx_a, a in enumerate(_ann["PVC_indices"]):
            if abs(r - a) < bias_thr:
                found = True
                beat_ann[idx_r] = "V"
                del _ann["PVC_indices"][idx_a]
                break
    ann_matched = {
        k: np.array([a for a in v if a not in _ann[k]], dtype=int)
        for k, v in ann.items()
    }
    retval = dict(ann_matched=ann_matched, beat_ann=beat_ann)
    return retval
    # _ann["SPB_indices"] = [a for a in _ann["SPB_indices"] if prev_r<a<next_r]
    # _ann["PVC_indices"] = [a for a in _ann["PVC_indices"] if prev_r<a<next_r]
    # augmented_rpeaks = np.concatenate((rpeaks, np.array(_ann["SPB_indices"]), np.array(_ann["PVC_indices"])))
    # beat_ann = np.concatenate((beat_ann, np.array(["S" for _ in _ann["SPB_indices"]], dtype="<U1"), np.array(["V" for _ in _ann["PVC_indices"]], dtype="<U1")))
    # sorted_indices = np.argsort(augmented_rpeaks)
    # augmented_rpeaks = augmented_rpeaks[sorted_indices].astype(int)
    # beat_ann = beat_ann[sorted_indices].astype("<U1")

    # retval = dict(augmented_rpeaks=augmented_rpeaks, beat_ann=beat_ann)
    # return retval


def _ann_to_beat_ann_epoch_v3(
    rpeaks: np.ndarray, ann: Dict[str, np.ndarray], bias_thr: Real
) -> dict:
    """

    similar to `_ann_to_beat_ann_epoch_v2`, but more reasonable

    Parameters
    ----------
    rpeaks: ndarray,
        rpeaks for forming beats
    ann: dict,
        with items (ndarray) "SPB_indices" and "PVC_indices",
        which record the indices of SPBs and PVCs
    bias_thr: real number,
        tolerance for using annotations (PVC, SPB indices provided by the dataset),
        to label the type of beats given by `rpeaks`

    Returns
    -------
    retval: dict, with the following items
        - ann_matched: dict of ndarray,
            indices of annotations ("SPB_indices" and "PVC_indices")
            that match some beat from `rpeaks`
        - beat_ann: ndarray,
            label for each beat from `rpeaks`

    """
    beat_ann = np.array(["N" for _ in range(len(rpeaks))], dtype="<U1")
    ann_matched = {k: [] for k, v in ann.items()}
    for idx_r, r in enumerate(rpeaks):
        dist_to_spb = np.abs(r - ann["SPB_indices"])
        dist_to_pvc = np.abs(r - ann["PVC_indices"])
        if len(dist_to_spb) == 0:
            dist_to_spb = np.array([np.inf])
        if len(dist_to_pvc) == 0:
            dist_to_pvc = np.array([np.inf])
        argmin = np.argmin([np.min(dist_to_spb), np.min(dist_to_pvc), bias_thr])
        if argmin == 2:
            pass
        elif argmin == 1:
            beat_ann[idx_r] = "V"
            ann_matched["PVC_indices"].append(
                ann["PVC_indices"][np.argmin(dist_to_pvc)]
            )
        elif argmin == 0:
            beat_ann[idx_r] = "S"
            ann_matched["SPB_indices"].append(
                ann["SPB_indices"][np.argmin(dist_to_spb)]
            )
    ann_matched = {k: np.array(v) for k, v in ann_matched.items()}
    retval = dict(ann_matched=ann_matched, beat_ann=beat_ann)
    return retval

    @property
    def url(self) -> str:
        return (
            "https://opensz.oss-cn-beijing.aliyuncs.com/ICBEB2020/file/TrainingSet.zip"
        )

    def download(self) -> NoReturn:
        """download the database from self.url"""
        http_get(self.url, self.db_dir, extract=True)


def compute_metrics(
    sbp_true: List[np.ndarray],
    pvc_true: List[np.ndarray],
    sbp_pred: List[np.ndarray],
    pvc_pred: List[np.ndarray],
    verbose: int = 0,
) -> Union[Tuple[int], dict]:
    """

    Score Function for all (test) records

    Parameters
    ----------
    sbp_true, pvc_true, sbp_pred, pvc_pred: list of ndarray,
    verbose: int

    Returns
    -------
    retval: tuple or dict,
        tuple of (negative) scores for each ectopic beat type (SBP, PVC), or
        dict of more scoring details, including
        - total_loss: sum of loss of each ectopic beat type (PVC and SPB)
        - true_positive: number of true positives of each ectopic beat type
        - false_positive: number of false positives of each ectopic beat type
        - false_negative: number of false negatives of each ectopic beat type

    """
    BaseCfg = CFG()
    BaseCfg.fs = 400
    BaseCfg.bias_thr = 0.15 * BaseCfg.fs
    s_score = np.zeros(
        [
            len(sbp_true),
        ],
        dtype=int,
    )
    v_score = np.zeros(
        [
            len(sbp_true),
        ],
        dtype=int,
    )
    # Scoring
    for i, (s_ref, v_ref, s_pos, v_pos) in enumerate(
        zip(sbp_true, pvc_true, sbp_pred, pvc_pred)
    ):
        s_tp = 0
        s_fp = 0
        s_fn = 0
        v_tp = 0
        v_fp = 0
        v_fn = 0
        # SBP
        if s_ref.size == 0:
            s_fp = len(s_pos)
        else:
            for m, ans in enumerate(s_ref):
                s_pos_cand = np.where(abs(s_pos - ans) <= BaseCfg.bias_thr)[0]
                if s_pos_cand.size == 0:
                    s_fn += 1
                else:
                    s_tp += 1
                    s_fp += len(s_pos_cand) - 1
        # PVC
        if v_ref.size == 0:
            v_fp = len(v_pos)
        else:
            for m, ans in enumerate(v_ref):
                v_pos_cand = np.where(abs(v_pos - ans) <= BaseCfg.bias_thr)[0]
                if v_pos_cand.size == 0:
                    v_fn += 1
                else:
                    v_tp += 1
                    v_fp += len(v_pos_cand) - 1
        # calculate the score
        s_score[i] = s_fp * (-1) + s_fn * (-5)
        v_score[i] = v_fp * (-1) + v_fn * (-5)

        if verbose >= 1:
            print(f"for the {i}-th record")
            print(f"s_tp = {s_tp}, s_fp = {s_fp}, s_fn = {s_fn}")
            print(f"v_tp = {v_tp}, v_fp = {v_fp}, s_fn = {v_fn}")
            print(f"s_score[{i}] = {s_score[i]}, v_score[{i}] = {v_score[i]}")

    Score1 = np.sum(s_score)
    Score2 = np.sum(v_score)

    if verbose >= 1:
        retval = CFG(
            total_loss=-(Score1 + Score2),
            class_loss={"S": -Score1, "V": -Score2},
            true_positive={"S": s_tp, "V": v_tp},
            false_positive={"S": s_fp, "V": v_fp},
            false_negative={"S": s_fn, "V": v_fn},
        )
    else:
        retval = Score1, Score2

    return retval



"""
the most basic CNN
"""

from copy import deepcopy
from typing import NoReturn, Optional, Sequence, Union

import torch
from torch import nn

from ...cfg import CFG, DEFAULTS
from ...models._nets import (  # noqa: F401
    Conv_Bn_Activation,
    GlobalContextBlock,
    NonLocalBlock,
    SEBlock,
)
from ...utils.misc import dict_to_str, add_docstring
from ...utils.utils_nn import (
    SizeMixin,
    compute_maxpool_output_shape,
    compute_sequential_output_shape,
    compute_sequential_output_shape_docstring,
)

if DEFAULTS.torch_dtype == torch.float64:
    torch.set_default_tensor_type(torch.DoubleTensor)


__all__ = [
    "VGGBlock",
    "VGG16",
]


class VGGBlock(nn.Sequential, SizeMixin):
    """building blocks of the CNN feature extractor `VGG16`"""

    __DEBUG__ = False
    __name__ = "VGGBlock"

    def __init__(
        self,
        num_convs: int,
        in_channels: int,
        out_channels: int,
        groups: int = 1,
        **config,
    ) -> NoReturn:
        """

        Parameters
        ----------
        num_convs: int,
            number of convolutional layers
        in_channels: int,
            number of channels in the input
        out_channels: int,
            number of channels produced by the convolutional layers
        groups: int, default 1,
            connection pattern (of channels) of the inputs and outputs
        config: dict,
            other parameters, including
            filter length (kernel size), activation choices,
            weight initializer, batch normalization choices, etc. for the convolutional layers;
            and pool size for the pooling layer

        """
        super().__init__()
        self.__num_convs = num_convs
        self.__in_channels = in_channels
        self.__out_channels = out_channels
        self.__groups = groups
        self.config = CFG(deepcopy(config))
        if self.__DEBUG__:
            print(
                f"configuration of {self.__name__} is as follows\n{dict_to_str(self.config)}"
            )

        self.add_module(
            "cba_1",
            Conv_Bn_Activation(
                in_channels,
                out_channels,
                kernel_size=self.config.filter_length,
                stride=self.config.subsample_length,
                groups=self.__groups,
                activation=self.config.activation,
                kw_activation=self.config.kw_activation,
                kernel_initializer=self.config.kernel_initializer,
                kw_initializer=self.config.kw_initializer,
                norm=self.config.get("norm", self.config.get("batch_norm")),
            ),
        )
        for idx in range(num_convs - 1):
            self.add_module(
                f"cba_{idx+2}",
                Conv_Bn_Activation(
                    out_channels,
                    out_channels,
                    kernel_size=self.config.filter_length,
                    stride=self.config.subsample_length,
                    groups=self.__groups,
                    activation=self.config.activation,
                    kw_activation=self.config.kw_activation,
                    kernel_initializer=self.config.kernel_initializer,
                    kw_initializer=self.config.kw_initializer,
                    norm=self.config.get("norm", self.config.get("batch_norm")),
                ),
            )
        self.add_module(
            "max_pool", nn.MaxPool1d(self.config.pool_size, self.config.pool_stride)
        )

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """

        Parameters
        ----------
        seq_len: int,
            length of the 1d sequence
        batch_size: int, optional,
            the batch size, can be None

        Returns
        -------
        output_shape: sequence,
            the output shape, given `seq_len` and `batch_size`
        """
        num_layers = 0
        for module in self:
            if num_layers < self.__num_convs:
                output_shape = module.compute_output_shape(seq_len, batch_size)
                _, _, seq_len = output_shape
            else:
                output_shape = compute_maxpool_output_shape(
                    input_shape=[batch_size, self.__out_channels, seq_len],
                    kernel_size=self.config.pool_size,
                    stride=self.config.pool_size,
                    channel_last=False,
                )
            num_layers += 1
        return output_shape


class VGG16(nn.Sequential, SizeMixin):
    """

    CNN feature extractor of the CRNN models proposed in refs of `ECG_CRNN`
    """

    __DEBUG__ = False
    __name__ = "VGG16"

    def __init__(self, in_channels: int, **config) -> NoReturn:
        """

        Parameters
        ----------
        in_channels: int,
            number of channels in the input
        config: dict,
            other hyper-parameters of the Module, including
            number of convolutional layers, number of filters for each layer,
            and more for `VGGBlock`.
            key word arguments that have to be set:
            num_convs: sequence of int,
                number of convolutional layers for each `VGGBlock`
            num_filters: sequence of int,
                number of filters for each `VGGBlock`
            groups: int,
                connection pattern (of channels) of the inputs and outputs
            block: dict,
                other parameters that can be set for `VGGBlock`
            for a full list of configurable parameters, ref. corr. config file

        """
        super().__init__()
        self.__in_channels = in_channels
        # self.config = deepcopy(ECG_CRNN_CONFIG.cnn.vgg16)
        self.config = CFG(deepcopy(config))
        if self.__DEBUG__:
            print(
                f"configuration of {self.__name__} is as follows\n{dict_to_str(self.config)}"
            )

        module_in_channels = in_channels
        for idx, (nc, nf) in enumerate(
            zip(self.config.num_convs, self.config.num_filters)
        ):
            module_name = f"vgg_block_{idx+1}"
            self.add_module(
                name=module_name,
                module=VGGBlock(
                    num_convs=nc,
                    in_channels=module_in_channels,
                    out_channels=nf,
                    groups=self.config.groups,
                    **(self.config.block),
                ),
            )
            module_in_channels = nf

    @add_docstring(compute_sequential_output_shape_docstring)
    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """ """
        return compute_sequential_output_shape(self, seq_len, batch_size)

    @property
    def in_channels(self) -> int:
        return self.__in_channels



"""
"""

from copy import deepcopy
from numbers import Real
from random import shuffle
from typing import Any, List, NoReturn, Optional, Sequence, Tuple

import numpy as np
import torch
from torch import Tensor

from .base import Augmenter
from ..cfg import DEFAULTS

__all__ = [
    "Mixup",
]


class Mixup(Augmenter):
    """
    Mixup augmentor.

    References
    ----------
    1. Zhang, Hongyi, et al. "mixup: Beyond Empirical Risk Minimization." International Conference on Learning Representations. 2018.
    2. https://arxiv.org/abs/1710.09412
    3. https://github.com/facebookresearch/mixup-cifar10/blob/master/train.py
    """

    __name__ = "Mixup"

    def __init__(
        self,
        fs: Optional[int] = None,
        alpha: Real = 0.5,
        beta: Optional[Real] = None,
        prob: float = 0.5,
        inplace: bool = True,
        **kwargs: Any
    ) -> NoReturn:
        """

        Parameters
        ----------
        fs: int, optional,
            sampling frequency of the ECGs to be augmented
        alpha: real number, default 0.5,
            alpha parameter of the Beta distribution used in Mixup.
        beta: real number, optional,
            beta parameter of the Beta distribution used in Mixup,
            default to alpha.
        prob: float, default 0.5,
            probability of applying Mixup.
        inplace: bool, default True,
            if True, ECG signal tensors will be modified inplace
        kwargs: Keyword arguments.
        """
        super().__init__()
        self.fs = fs
        self.alpha = alpha
        self.beta = beta or self.alpha
        self.prob = prob
        assert 0 <= self.prob <= 1, "Probability must be between 0 and 1"
        self.inplace = inplace

    def forward(
        self,
        sig: Tensor,
        label: Tensor,
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """

        Parameters
        ----------
        sig: Tensor,
            the ECGs to be augmented, of shape (batch, lead, siglen)
        label: Tensor,
            label tensor of the ECGs
        extra_tensors: sequence of Tensors, optional,
            not used, but kept for consistency with other augmenters
        kwargs: keyword arguments,
            not used, but kept for consistency with other augmenters

        Returns
        -------
        sig: Tensor,
            the augmented ECGs
        label: Tensor,
            the augmented label
        extra_tensors: sequence of Tensors, optional,
            if set in the input arguments, unchanged
        """
        batch, lead, siglen = sig.shape
        # TODO: make `lam` different for each batch element, using
        # lam = torch.from_numpy(DEFAULTS.RNG.beta(self.alpha, self.beta, batch), dtype=sig.dtype, device=sig.device)
        # of shape (batch,)
        lam = DEFAULTS.RNG.beta(self.alpha, self.beta)
        indices = np.arange(batch, dtype=int)
        ori = self.get_indices(prob=self.prob, pop_size=batch)
        # print(f"ori = {ori}, len(ori) = {len(ori)}")
        perm = deepcopy(ori)
        shuffle(perm)
        indices[ori] = perm
        indices = torch.from_numpy(indices).long()

        if not self.inplace:
            sig = sig.clone()
            label = label.clone()

        sig = lam * sig + (1 - lam) * sig[indices]
        label = lam * label + (1 - lam) * label[indices]
        # TODO: if `lam` is a Tensor of shape (batch,) instead of a scalar,
        # sig = lam.view(batch, 1, 1) * sig + (1 - lam).view(batch, 1, 1) * sig[indices]
        # label = lam.view(batch, 1) * label + (1 - lam).view(batch, 1) * label[indices]

        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        """ """
        return [
            "alpha",
            "beta",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



"""
"""

from numbers import Real
from typing import Any, List, NoReturn, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from .base import Augmenter

__all__ = [
    "RandomFlip",
]


class RandomFlip(Augmenter):
    """ """

    __name__ = "RandomFlip"

    def __init__(
        self,
        fs: Optional[int] = None,
        per_channel: bool = True,
        prob: Union[Sequence[float], float] = [0.4, 0.2],
        inplace: bool = True,
        **kwargs: Any
    ) -> NoReturn:
        """

        Parameters
        ----------
        fs: int, optional,
            sampling frequency of the ECGs to be augmented
        per_channel: bool, default True,
            whether to flip each channel independently.
        prob: sequence of float or float, default [0.4,0.2],
            probability of performing flip,
            the first probality is for the batch dimension,
            the second probability is for the lead dimension.
        inplace: bool, default True,
            if True, ECG signal tensors will be modified inplace
        kwargs: Keyword arguments.
        """
        super().__init__()
        self.fs = fs
        self.per_channel = per_channel
        self.inplace = inplace
        self.prob = prob
        if isinstance(self.prob, Real):
            self.prob = np.array([self.prob, self.prob])
        else:
            self.prob = np.array(self.prob)
        assert (self.prob >= 0).all() and (
            self.prob <= 1
        ).all(), "Probability must be between 0 and 1"

    def forward(
        self,
        sig: Tensor,
        label: Optional[Tensor],
        *extra_tensors: Sequence[Tensor],
        **kwargs: Any
    ) -> Tuple[Tensor, ...]:
        """

        Parameters
        ----------
        sig: Tensor,
            the ECGs to be augmented, of shape (batch, lead, siglen)
        label: Tensor, optional,
            label tensor of the ECGs,
            not used, but kept for consistency with other augmenters
        extra_tensors: sequence of Tensors, optional,
            not used, but kept for consistency with other augmenters
        kwargs: keyword arguments,
            not used, but kept for consistency with other augmenters

        Returns
        -------
        sig: Tensor,
            the augmented ECGs
        label: Tensor,
            the label tensor of the augmented ECGs, unchanged
        extra_tensors: sequence of Tensors, optional,
            if set in the input arguments, unchanged
        """
        batch, lead, siglen = sig.shape
        if not self.inplace:
            sig = sig.clone()
        if self.prob[0] == 0:
            return (sig, label, *extra_tensors)
        if self.per_channel:
            flip = torch.ones((batch, lead, 1), dtype=sig.dtype, device=sig.device)
            for i in self.get_indices(prob=self.prob[0], pop_size=batch):
                flip[i, self.get_indices(prob=self.prob[1], pop_size=lead), ...] = -1
            sig = sig.mul_(flip)
        else:
            flip = torch.ones((batch, 1, 1), dtype=sig.dtype, device=sig.device)
            flip[self.get_indices(prob=self.prob[0], pop_size=batch), ...] = -1
            sig = sig.mul_(flip)
        return (sig, label, *extra_tensors)

    def extra_repr_keys(self) -> List[str]:
        """ """
        return [
            "per_channel",
            "prob",
            "inplace",
        ] + super().extra_repr_keys()



"""
"""

from random import sample
from typing import List, NoReturn, Optional, Tuple

import torch
import torch.nn as nn

from ..utils.misc import ReprMixin
from .bandpass import BandPass
from .baseline_remove import BaselineRemove
from .normalize import Normalize
from .resample import Resample

__all__ = [
    "PreprocManager",
]


class PreprocManager(ReprMixin, nn.Module):
    """

    Examples
    --------
    ```python
    import torch
    from torch_ecg.cfg import CFG
    from torch_ecg.preprocessors import PreprocManager

    config = CFG(
        random=False,
        bandpass={"fs":500},
        normalize={"method": "min-max"},
    )
    ppm = PreprocManager.from_config(config)
    sig = torch.rand(2,12,8000)
    sig = ppm(sig)
    ```

    """

    __name__ = "PreprocManager"

    def __init__(
        self,
        *pps: Optional[Tuple[nn.Module, ...]],
        random: bool = False,
        inplace: bool = True,
    ) -> NoReturn:
        """

        Parameters
        ----------
        pps: tuple of `nn.Module`, optional,
            the sequence of preprocessors to be added to the manager
        random: bool, default False,
            whether to apply the augmenters in random order
        inplace: bool, default True,
            whether to apply the preprocessors in-place

        """
        super().__init__()
        self.random = random
        self._preprocessors = list(pps)

    def _add_bandpass(self, **config: dict) -> NoReturn:
        """ """
        self._preprocessors.append(BandPass(**config))

    def _add_baseline_remove(self, **config: dict) -> NoReturn:
        """ """
        self._preprocessors.append(BaselineRemove(**config))

    def _add_normalize(self, **config: dict) -> NoReturn:
        """ """
        self._preprocessors.append(Normalize(**config))

    def _add_resample(self, **config: dict) -> NoReturn:
        """ """
        self._preprocessors.append(Resample(**config))

    def forward(self, sig: torch.Tensor) -> torch.Tensor:
        """

        Parameters
        ----------
        sig: Tensor,
            the signal tensor to be preprocessed

        Returns
        -------
        sig: Tensor,
            the preprocessed signal tensor

        """
        if len(self.preprocessors) == 0:
            raise ValueError("No preprocessors added to the manager.")
        ordering = list(range(len(self.preprocessors)))
        if self.random:
            ordering = sample(ordering, len(ordering))
        for idx in ordering:
            sig = self.preprocessors[idx](sig)
        return sig

    @classmethod
    def from_config(cls, config: dict) -> "PreprocManager":
        """

        Parameters
        ----------
        config: dict,
            the configuration of the preprocessors,
            better to be an `OrderedDict`

        Returns
        -------
        ppm: PreprocManager,
            a new instance of `PreprocManager`

        """
        ppm = cls(
            random=config.get("random", False), inplace=config.get("inplace", True)
        )
        _mapping = {
            "bandpass": ppm._add_bandpass,
            "baseline_remove": ppm._add_baseline_remove,
            "normalize": ppm._add_normalize,
            "resample": ppm._add_resample,
        }
        for pp_name, pp_config in config.items():
            if pp_name in [
                "random",
                "inplace",
            ]:
                continue
            if pp_name in _mapping and isinstance(pp_config, dict):
                _mapping[pp_name](**pp_config)
            else:
                # just ignore the other items
                pass
                # raise ValueError(f"Unknown preprocessor: {pp_name}")
        return ppm

    def rearrange(self, new_ordering: List[str]) -> NoReturn:
        """

        Parameters
        ----------
        new_ordering: list of str,
            the new ordering of the preprocessors

        """
        _mapping = {
            "Resample": "resample",
            "BandPass": "bandpass",
            "BaselineRemove": "baseline_remove",
            "Normalize": "normalize",
        }
        for k in new_ordering:
            if k not in _mapping:
                _mapping.update({k: k})
        self._preprocessors.sort(
            key=lambda aug: new_ordering.index(_mapping[aug.__name__])
        )

    def add_(self, pp: nn.Module, pos: int = -1) -> NoReturn:
        """

        add a (custom) preprocessor to the manager,
        this method is preferred against directly manipulating
        the internal list of preprocessors via `PreprocManager.preprocessors.append(pp)`

        Parameters
        ----------
        pp: Module,
            the preprocessor to be added
        pos: int, default -1,
            the position to insert the preprocessor,
            should be >= -1, with -1 the indicator of the end

        """
        assert isinstance(pp, nn.Module)
        assert pp.__class__.__name__ not in [
            p.__class__.__name__ for p in self.preprocessors
        ], f"Preprocessor {pp.__class__.__name__} already exists."
        assert (
            isinstance(pos, int) and pos >= -1
        ), f"pos must be an integer >= -1, but got {pos}."
        if pos == -1:
            self._preprocessors.append(pp)
        else:
            self._preprocessors.insert(pos, pp)

    @property
    def preprocessors(self) -> List[nn.Module]:
        return self._preprocessors

    def extra_repr_keys(self) -> List[str]:
        """
        return the extra keys for `__repr__`

        """
        return super().extra_repr_keys() + [
            "random",
            "preprocessors",
        ]
